@@ -39,16 +39,19 @@ extern PSI_memory_key key_memory_String_value;
 
 typedef struct st_io_cache IO_CACHE;
 typedef struct st_mem_root MEM_ROOT;
+#define ASSERT_LENGTH(A) DBUG_ASSERT(str_length + (uint32) (A) <= Alloced_length)
 
 #include "pack.h"
-int sortcmp(const String *a,const String *b, CHARSET_INFO *cs);
+class Binary_string;
+int sortcmp(const Binary_string *s, const Binary_string *t, CHARSET_INFO *cs);
+int stringcmp(const Binary_string *s, const Binary_string *t);
 String *copy_if_not_alloced(String *a,String *b,uint32 arg_length);
-inline uint32 copy_and_convert(char *to, size_t to_length,
-                               CHARSET_INFO *to_cs,
+inline uint32 copy_and_convert(char *to, size_t to_length, CHARSET_INFO *to_cs,
                                const char *from, size_t from_length,
                                CHARSET_INFO *from_cs, uint *errors)
 {
-  return my_convert(to, (uint)to_length, to_cs, from, (uint)from_length, from_cs, errors);
+  return my_convert(to, (uint)to_length, to_cs, from, (uint)from_length,
+                    from_cs, errors);
 }
 
 
@@ -110,7 +113,8 @@ public:
     "dstcs" and "srccs" cannot be &my_charset_bin.
   */
   size_t convert_fix(CHARSET_INFO *dstcs, char *dst, size_t dst_length,
-                     CHARSET_INFO *srccs, const char *src, size_t src_length, size_t nchars)
+                     CHARSET_INFO *srccs, const char *src, size_t src_length,
+                     size_t nchars)
   {
     return my_convert_fix(dstcs, dst, dst_length,
                           srccs, src, src_length, nchars, this, this);
@@ -119,10 +123,12 @@ public:
      Copy a string. Fix bad bytes/characters to '?'.
   */
   uint well_formed_copy(CHARSET_INFO *to_cs, char *to, size_t to_length,
-                        CHARSET_INFO *from_cs, const char *from, size_t from_length, size_t nchars);
+                        CHARSET_INFO *from_cs, const char *from,
+                        size_t from_length, size_t nchars);
   // Same as above, but without the "nchars" limit.
   uint well_formed_copy(CHARSET_INFO *to_cs, char *to, size_t to_length,
-                        CHARSET_INFO *from_cs, const char *from, size_t from_length)
+                        CHARSET_INFO *from_cs, const char *from,
+                        size_t from_length)
   {
     return well_formed_copy(to_cs, to, to_length,
                             from_cs, from, from_length,
@@ -183,7 +189,7 @@ public:
   }
   bool same_encoding(const Charset &other) const
   {
-    return !strcmp(m_charset->csname, other.m_charset->csname);
+    return my_charset_same(m_charset, other.m_charset);
   }
   /*
     Collation name without the character set name.
@@ -277,26 +283,75 @@ public:
 };
 
 
-/*
-  A storage for String.
-  Should be eventually derived from LEX_STRING.
+/**
+   Storage for strings with both length and allocated length.
+   Automatically grows on demand.
 */
-class Static_binary_string : public Sql_alloc
+
+class Binary_string: public Sql_alloc
 {
 protected:
   char *Ptr;
-  uint32 str_length;
-public:
-  Static_binary_string()
-   :Ptr(NULL),
-    str_length(0)
-  { }
-  Static_binary_string(char *str, size_t length_arg)
-   :Ptr(str),
-    str_length((uint32) length_arg)
+  uint32 str_length, Alloced_length, extra_alloc;
+  bool alloced, thread_specific;
+  void init_private_data()
   {
-    DBUG_ASSERT(length_arg < UINT_MAX32);
+    Ptr= 0;
+    Alloced_length= extra_alloc= str_length= 0;
+    alloced= thread_specific= false;
   }
+  inline void free_buffer()
+  {
+    if (alloced)
+    {
+      alloced=0;
+      my_free(Ptr);
+    }
+  }
+public:
+  Binary_string()
+  {
+    init_private_data();
+  }
+  explicit Binary_string(size_t length_arg)
+  {
+    init_private_data();
+    (void) real_alloc(length_arg);
+  }
+  /*
+    NOTE: If one intend to use the c_ptr() method, the following two
+    contructors need the size of memory for STR to be at least LEN+1 (to make
+    room for zero termination).
+  */
+  Binary_string(const char *str, size_t len)
+  {
+    Ptr= (char*) str;
+    str_length= (uint32) len;
+    Alloced_length= 0;                          /* Memory cannot be written to */
+    extra_alloc= 0;
+    alloced= thread_specific= 0;
+  }
+  Binary_string(char *str, size_t len)
+  {
+    Ptr= str;
+    str_length= Alloced_length= (uint32) len;
+    extra_alloc= 0;
+    alloced= thread_specific= 0;
+  }
+  explicit Binary_string(const Binary_string &str)
+  {
+    Ptr= str.Ptr;
+    str_length= str.str_length;
+    Alloced_length= str.Alloced_length;
+    extra_alloc= 0;
+    alloced= thread_specific= 0;
+  }
+
+  ~Binary_string()
+  {
+    free();
+  }
+
   inline uint32 length() const { return str_length;}
   inline char& operator [] (size_t i) const { return Ptr[i]; }
   inline void length(size_t len) { str_length=(uint32)len ; }
@@ -313,24 +368,12 @@ public:
     return false;
   }
 
-  bool bin_eq(const Static_binary_string *other) const
+  bool bin_eq(const Binary_string *other) const
   {
     return length() == other->length() &&
            !memcmp(ptr(), other->ptr(), length());
   }
 
-  void set(char *str, size_t len)
-  {
-    Ptr= str;
-    str_length= (uint32) len;
-  }
-
-  void swap(Static_binary_string &s)
-  {
-    swap_variables(char *, Ptr, s.Ptr);
-    swap_variables(uint32, str_length, s.str_length);
-  }
-
   /*
     PMG 2004.11.12
     This is a method that works the same as perl's "chop". It simply
@@ -354,42 +397,52 @@ public:
   */
   inline void chop()
   {
-    str_length--;
-    Ptr[str_length]= '\0';
-    DBUG_ASSERT(strlen(Ptr) == str_length);
+    if (str_length)
+    {
+      str_length--;
+      Ptr[str_length]= '\0';
+      DBUG_ASSERT(strlen(Ptr) == str_length);
+    }
   }
 
   // Returns offset to substring or -1
-  int strstr(const Static_binary_string &search, uint32 offset=0);
+  int strstr(const Binary_string &search, uint32 offset=0) const;
+  int strstr(const char *search, uint32 search_length, uint32 offset=0) const;
   // Returns offset to substring or -1
-  int strrstr(const Static_binary_string &search, uint32 offset=0);
+  int strrstr(const Binary_string &search, uint32 offset=0) const;
 
   /*
-    The following append operations do NOT check alloced memory
+    The following append operations do not extend the strings and in production
+    mode do NOT check that alloced memory!
     q_*** methods writes values of parameters itself
     qs_*** methods writes string representation of value
   */
   void q_append(const char c)
   {
+    ASSERT_LENGTH(1);
     Ptr[str_length++] = c;
   }
   void q_append2b(const uint32 n)
   {
+    ASSERT_LENGTH(2);
     int2store(Ptr + str_length, n);
     str_length += 2;
   }
   void q_append(const uint32 n)
   {
+    ASSERT_LENGTH(4);
     int4store(Ptr + str_length, n);
     str_length += 4;
   }
   void q_append(double d)
   {
+    ASSERT_LENGTH(8);
     float8store(Ptr + str_length, d);
     str_length += 8;
   }
   void q_append(double *d)
   {
+    ASSERT_LENGTH(8);
     float8store(Ptr + str_length, *d);
     str_length += 8;
   }
@@ -408,6 +461,7 @@ public:
   }
   void q_append(const char *data, size_t data_len)
   {
+    ASSERT_LENGTH(data_len);
     if (data_len)
       memcpy(Ptr + str_length, data, data_len);
     DBUG_ASSERT(str_length <= UINT_MAX32 - data_len);
@@ -421,15 +475,12 @@ public:
     q_append(ls->str, (uint32) ls->length);
   }
 
-  void write_at_position(int position, uint32 value)
+  void write_at_position(uint32 position, uint32 value)
   {
+    DBUG_ASSERT(str_length >= position + 4);
     int4store(Ptr + position,value);
   }
 
-  void qs_append(const char *str)
-  {
-    qs_append(str, (uint32)strlen(str));
-  }
   void qs_append(const LEX_CSTRING *ls)
   {
     DBUG_ASSERT(ls->length < UINT_MAX32 &&
@@ -444,8 +495,9 @@ public:
   void qs_append(const double *d);
   inline void qs_append(const char c)
   {
-     Ptr[str_length]= c;
-     str_length++;
+    ASSERT_LENGTH(1);
+    Ptr[str_length]= c;
+    str_length++;
   }
   void qs_append(int i);
   void qs_append(uint i)
@@ -459,65 +511,11 @@ public:
   void qs_append(ulonglong i);
   void qs_append(longlong i, int radix)
   {
+    ASSERT_LENGTH(22);
     char *buff= Ptr + str_length;
     char *end= ll2str(i, buff, radix, 0);
-    str_length+= uint32(end-buff);
+    str_length+= (uint32) (end-buff);
   }
-};
-
-
-class Binary_string: public Static_binary_string
-{
-  uint32 Alloced_length, extra_alloc;
-  bool alloced, thread_specific;
-  void init_private_data()
-  {
-    Alloced_length= extra_alloc= 0;
-    alloced= thread_specific= false;
-  }
-public:
-  Binary_string()
-  {
-    init_private_data();
-  }
-  explicit Binary_string(size_t length_arg)
-  {
-    init_private_data();
-    (void) real_alloc(length_arg);
-  }
-  explicit Binary_string(const char *str)
-   :Binary_string(str, strlen(str))
-  { }
-  /*
-    NOTE: If one intend to use the c_ptr() method, the following two
-    contructors need the size of memory for STR to be at least LEN+1 (to make
-    room for zero termination).
-  */
-  Binary_string(const char *str, size_t len)
-   :Static_binary_string((char *) str, len)
-  {
-    init_private_data();
-  }
-  Binary_string(char *str, size_t len)
-   :Static_binary_string(str, len)
-  {
-    Alloced_length= (uint32) len;
-    extra_alloc= 0;
-    alloced= thread_specific= 0;
-  }
-  explicit Binary_string(const Binary_string &str)
-   :Static_binary_string(str)
-  {
-    Alloced_length= str.Alloced_length;
-    extra_alloc= 0;
-    alloced= thread_specific= 0;
-  }
-  Binary_string(Binary_string &&str) noexcept
-  {
-    move(str);
-  }
-
-  ~Binary_string() { free(); }
 
   /* Mark variable thread specific it it's not allocated already */
   inline void set_thread_specific()
@@ -533,13 +531,14 @@ public:
 
   inline bool uses_buffer_owned_by(const Binary_string *s) const
   {
-    return (s->alloced && Ptr >= s->Ptr && Ptr < s->Ptr + s->str_length);
+    return (s->alloced && Ptr >= s->Ptr && Ptr < s->Ptr + s->Alloced_length);
   }
 
   /* Swap two string objects. Efficient way to exchange data without memcpy. */
   void swap(Binary_string &s)
   {
-    Static_binary_string::swap(s);
+    swap_variables(char *, Ptr, s.Ptr);
+    swap_variables(uint32, str_length, s.str_length);
     swap_variables(uint32, Alloced_length, s.Alloced_length);
     swap_variables(bool, alloced, s.alloced);
   }
@@ -551,28 +550,33 @@ public:
             null character.
      @note The new buffer will not be null terminated.
   */
-  void set_alloced(char *str, size_t length_arg, size_t alloced_length_arg)
+  void set_alloced(char *str, size_t length, size_t alloced_length)
   {
-    free();
-    Static_binary_string::set(str, length_arg);
-    DBUG_ASSERT(alloced_length_arg < UINT_MAX32);
-    Alloced_length= (uint32) alloced_length_arg;
+    free_buffer();
+    Ptr= str;
+    str_length= (uint32) length;
+    DBUG_ASSERT(alloced_length < UINT_MAX32);
+    Alloced_length= (uint32) alloced_length;
   }
   inline void set(char *str, size_t arg_length)
   {
     set_alloced(str, arg_length, arg_length);
   }
-  inline void set(const char *str, size_t arg_length)
+  inline void set(const char *str, size_t length)
   {
-    free();
-    Static_binary_string::set((char *) str, arg_length);
+    free_buffer();
+    Ptr= (char*) str;
+    str_length= (uint32) length;
+    Alloced_length= 0;
   }
 
-  void set(Binary_string &str, size_t offset, size_t arg_length)
+  void set(Binary_string &str, size_t offset, size_t length)
   {
     DBUG_ASSERT(&str != this);
-    free();
-    Static_binary_string::set((char*) str.ptr() + offset, arg_length);
+    free_buffer();
+    Ptr= str.Ptr + offset;
+    str_length= (uint32) length;
+    Alloced_length= 0;
     if (str.Alloced_length)
       Alloced_length= (uint32) (str.Alloced_length - offset);
   }
@@ -599,18 +603,29 @@ public:
   char *release()
   {
     char *old= Ptr;
-    Static_binary_string::set(NULL, 0);
     init_private_data();
     return old;
   }
 
-  inline void set_quick(char *str, size_t arg_length)
+  /*
+    This is used to set a new buffer for String.
+    However if the String already has an allocated buffer, it will
+    keep that one.
+    It's not to be used to set the value or length of the string.
+  */
+  inline void set_buffer_if_not_allocated(char *str, size_t arg_length)
   {
     if (!alloced)
     {
-      Static_binary_string::set(str, arg_length);
-      Alloced_length= (uint32) arg_length;
+      /*
+        Following should really set str_length= 0, but some code may
+        depend on that the String length is same as buffer length.
+      */
+      Ptr= str;
+      str_length= Alloced_length= (uint32) arg_length;
     }
+    /* One should set str_length before using it */
+    MEM_UNDEFINED(&str_length, sizeof(str_length));
   }
 
   inline Binary_string& operator=(const Binary_string &s)
@@ -710,19 +725,49 @@ public:
 
   inline char *c_ptr()
   {
-    DBUG_ASSERT(!alloced || !Ptr || !Alloced_length ||
-                (Alloced_length >= (str_length + 1)));
+    if (unlikely(!Ptr))
+      return (char*) "";
+    /*
+      Here we assume that any buffer used to initalize String has
+      an end \0 or have at least an accessable character at end.
+      This is to handle the case of String("Hello",5) and
+      String("hello",5) efficiently.
 
-    if (!Ptr || Ptr[str_length])              // Should be safe
-      (void) realloc(str_length);
+      We have two options here. To test for !Alloced_length or !alloced.
+      Using "Alloced_length" is slightly safer so that we do not read
+      from potentially unintialized memory (normally not dangerous but
+      may give warnings in valgrind), but "alloced" is safer as there
+      are less change to get memory loss from code that is using
+      String((char*), length) or String.set((char*), length) and does
+      not free things properly (and there is several places in the code
+      where this happens and it is hard to find out if any of these will call
+      c_ptr().
+    */
+    if (unlikely(!alloced && !Ptr[str_length]))
+      return Ptr;
+    if (str_length < Alloced_length)
+    {
+      Ptr[str_length]=0;
+      return Ptr;
+    }
+    (void) realloc(str_length);               /* This will add end \0 */
     return Ptr;
   }
+  /*
+    One should use c_ptr() instead for most cases. This will be deleted soon,
+    kept for compatiblity.
+  */
   inline char *c_ptr_quick()
   {
-    if (Ptr && str_length < Alloced_length)
-      Ptr[str_length]=0;
-    return Ptr;
+    return c_ptr_safe();
   }
+  /*
+    This is to be used only in the case when one cannot use c_ptr().
+    The cases are:
+    - When one initializes String with an external buffer and length and
+      buffer[length] could be uninitalized when c_ptr() is called.
+    - When valgrind gives warnings about uninitialized memory with c_ptr().
+  */
   inline char *c_ptr_safe()
   {
     if (Ptr && str_length < Alloced_length)
@@ -734,17 +779,28 @@ public:
 
   inline void free()
   {
-    if (alloced)
-    {
-      alloced=0;
-      my_free(Ptr);
-    }
+    free_buffer();
+    /*
+      We have to clear the values as some Strings, like in Field, are
+      reused after free(). Because of this we cannot use MEM_UNDEFINED() here.
+    */
+    Ptr= 0;
+    str_length= 0;
     Alloced_length= extra_alloc= 0;
-    Static_binary_string::set(NULL, 0); // Safety
   }
+
   inline bool alloc(size_t arg_length)
   {
-    if (arg_length < Alloced_length)
+    /*
+      Allocate if we need more space or if we don't have done any
+      allocation yet (we don't want to have Ptr to be NULL for empty strings).
+
+      Note that if arg_length == Alloced_length then we don't allocate.
+      This ensures we don't do any extra allocations in protocol and String:int,
+      but the string will not be automatically null terminated if c_ptr() is not
+      called.
+    */
+    if (arg_length <= Alloced_length && Alloced_length)
       return 0;
     return real_alloc(arg_length);
   }
@@ -752,7 +808,7 @@ public:
   bool realloc_raw(size_t arg_length);
   bool realloc(size_t arg_length)
   {
-    if (realloc_raw(arg_length))
+    if (realloc_raw(arg_length+1))
       return TRUE;
     Ptr[arg_length]= 0; // This make other funcs shorter
     return FALSE;
@@ -786,13 +842,13 @@ public:
     thread_specific= s.thread_specific;
     s.alloced= 0;
   }
-  bool fill(uint32 max_length,char fill);
+  bool fill(size_t max_length,char fill);
   /*
     Replace substring with string
     If wrong parameter or not enough memory, do nothing
   */
   bool replace(uint32 offset,uint32 arg_length, const char *to, uint32 length);
-  bool replace(uint32 offset,uint32 arg_length, const Static_binary_string &to)
+  bool replace(uint32 offset,uint32 arg_length, const Binary_string &to)
   {
     return replace(offset,arg_length,to.ptr(),to.length());
   }
@@ -839,12 +895,7 @@ class String: public Charset, public Binary_string
 {
 public:
   String() = default;
-  String(size_t length_arg)
-   :Binary_string(length_arg)
-  { }
-  String(const char *str, CHARSET_INFO *cs)
-   :Charset(cs),
-    Binary_string(str)
+  String(size_t length_arg) :Binary_string(length_arg)
   { }
   /*
     NOTE: If one intend to use the c_ptr() method, the following two
@@ -852,16 +903,12 @@ public:
     room for zero termination).
   */
   String(const char *str, size_t len, CHARSET_INFO *cs)
-   :Charset(cs),
-    Binary_string((char *) str, len)
+   :Charset(cs), Binary_string(str, len)
   { }
   String(char *str, size_t len, CHARSET_INFO *cs)
-   :Charset(cs),
-    Binary_string(str, len)
+   :Charset(cs), Binary_string(str, len)
   { }
   String(const String &str) = default;
-  String(String &&str) noexcept
-   :Charset(std::move(str)), Binary_string(std::move(str)){}
 
   void set(String &str,size_t offset,size_t arg_length)
   {
@@ -879,9 +926,10 @@ public:
     set_charset(cs);
   }
   bool set_ascii(const char *str, size_t arg_length);
-  inline void set_quick(char *str,size_t arg_length, CHARSET_INFO *cs)
+  inline void set_buffer_if_not_allocated(char *str,size_t arg_length,
+                                          CHARSET_INFO *cs)
   {
-    Binary_string::set_quick(str, arg_length);
+    Binary_string::set_buffer_if_not_allocated(str, arg_length);
     set_charset(cs);
   }
   bool set_int(longlong num, bool unsigned_flag, CHARSET_INFO *cs);
@@ -977,7 +1025,8 @@ public:
     if (unlikely(alloc(tocs->mbmaxlen * src_length)))
       return true;
     str_length= copier->well_formed_copy(tocs, Ptr, alloced_length(),
-                                         fromcs, src, (uint)src_length, (uint)nchars);
+                                         fromcs, src, (uint) src_length,
+                                         (uint) nchars);
     set_charset(tocs);
     return false;
   }
@@ -1001,8 +1050,8 @@ public:
   bool append_introducer_and_hex(const String *str)
   {
     return
-      append(STRING_WITH_LEN("_"))   ||
-      append(str->charset()->csname)  ||
+      append('_')   ||
+      append(str->charset()->cs_name) ||
       append(STRING_WITH_LEN(" 0x")) ||
       append_hex(str->ptr(), (uint32) str->length());
   }
@@ -1016,17 +1065,6 @@ public:
   }
 
   // Append with optional character set conversion from ASCII (e.g. to UCS2)
-  bool append(const char *s)
-  {
-    return append(s, strlen(s));
-  }
-  bool append(const LEX_STRING *ls)
-  {
-    DBUG_ASSERT(ls->length < UINT_MAX32 &&
-                ((ls->length == 0 && !ls->str) ||
-                 ls->length == strlen(ls->str)));
-    return append(ls->str, (uint32) ls->length);
-  }
   bool append(const LEX_CSTRING *ls)
   {
     DBUG_ASSERT(ls->length < UINT_MAX32 &&
@@ -1093,7 +1131,7 @@ public:
     @param names  - an array of flag names
     @param count  - the number of available elements in "names"
   */
-  bool append_flag32_names(uint32 flags, const char *names[], size_t count)
+  bool append_flag32_names(uint32 flags, LEX_CSTRING names[], size_t count)
   {
     bool added= false;
     if (flags && append('('))
@@ -1105,7 +1143,7 @@ public:
       {
         if (added && append('|'))
           return true;
-        if (append(bit < count ? names[bit] : "?"))
+        if (bit < count ? append(names[bit]) : append('?'))
           return true;
         added= true;
       }
@@ -1116,8 +1154,6 @@ public:
   }
 
   void strip_sp();
-  friend int sortcmp(const String *a,const String *b, CHARSET_INFO *cs);
-  friend int stringcmp(const String *a,const String *b);
   friend String *copy_if_not_alloced(String *a,String *b,uint32 arg_length);
   friend class Field;
   uint32 numchars() const
@@ -1147,11 +1183,12 @@ public:
 
   static my_wc_t escaped_wc_for_single_quote(my_wc_t ch)
   {
-    switch (ch)
-    {
+    switch (ch) {
     case '\\':   return '\\';
     case '\0':   return '0';
     case '\'':   return '\'';
+    case '\b':   return 'b';
+    case '\t':   return 't';
     case '\n':   return 'n';
     case '\r':   return 'r';
     case '\032': return 'Z';
@@ -1186,12 +1223,6 @@ public:
   {
     return append_for_single_quote(s->ptr(), s->length());
   }
-  bool append_for_single_quote(const char *st)
-  {
-    size_t len= strlen(st);
-    DBUG_ASSERT(len < UINT_MAX32);
-    return append_for_single_quote(st, (uint32) len);
-  }
 
   void swap(String &s)
   {
@@ -1243,6 +1274,18 @@ public:
   {
     length(0);
   }
+  void set_buffer_if_not_allocated(CHARSET_INFO *cs)
+  {
+    if (!is_alloced())
+    {
+      Ptr= buff;
+      Alloced_length= (uint32) buff_sz;
+    }
+    str_length= 0;                          /* Safety, not required */
+    /* One should set str_length before using it */
+    MEM_UNDEFINED(&str_length, sizeof(str_length));
+    set_charset(cs);
+  }
 };
 
 
@@ -1254,18 +1297,6 @@ public:
   BinaryStringBuffer() : Binary_string(buff, buff_sz) { length(0); }
 };
 
-
-class String_space: public String
-{
-public:
-  String_space(uint n)
-  {
-    if (fill(n, ' '))
-      set("", 0, &my_charset_bin);
-  }
-};
-
-
 static inline bool check_if_only_end_space(CHARSET_INFO *cs,
                                            const char *str,
                                            const char *end)
